import re, os, sys, time, queue, traceback
from threading import Thread
from engine.as_sqlite import asSqlite
from engine.as_sqlite import asSqlite_deferred
from engine.as_http import asHttp
from engine.as_ac import asAc
from engine.as_hlp import hlp

class asBot():
	''' базовый класс для модуля '''

	data = {}

	def __init__(self, parent):
		self.parent = parent

		to_global = ['log', 'gui', 'file', 'http', 'queue',
					 'proxy', 'cnf', 'threads', 'sqlite', 'hlp']

		self.hlp = hlp(self.parent)
		self.data = self.parent.data

		for mod in to_global:
			setattr(self, mod, getattr(self.parent, mod))

		with open('engine/cmd.txt', 'wb') as f:
			pass # clear file

	def init(self):
		''' инициализация бота и настройка gui '''

		# новый поток чекающий engine/cmd.txt  И  обновляющий счётчики
		try:
			t2 = Thread(target=self.w_check_cmd)
			t2.start()
		except Exception as e:
			print('include/as_bot: self.w_check_cmd EXCEPTION: %s' % e)

		# поток, пополняющий parent.sqlite_storage
		#~ t3 = Thread(target=self.w_sqlite_cache)
		#~ t3.start()

		# поток, выполняющий команды к sqlite
		t4 = Thread(target=self.w_sqlite)
		t4.start()

		# поток, собирающий сообщения для записи в лог
		t5 = Thread(target=self.w_text_log)
		t5.start()

		# поток обновляющий прокси - находится в as_proxy.py
		# w_autoupdate

	def preload_tables(self, tables=[]):
		# принимает список таблиц ['*accounts', 'validol']
		# кеширует их данные в self.data
		# возвращает True/False

		if isinstance(tables, str):
			tables = [tables]

		result_done = True
		for table in tables:

			critical = False
			if table.startswith('*'):
				critical = True
				table = table[1:]

			self.data[table] = self.sqlite.getlist(table)
			sz = len(self.data[table])
			self.log('из "{0}" загружено {1} строк'.format(table, sz))

			if critical and not sz:
				self.log('нет данных в таблице "{0}"'.format(sz), '!')
				result_done = False

		return result_done

	def preload_files_dirs(self, files):
		# принимает список файлов и папок ['*accounts', 'links', 'videos/']
		# кеширует их данные в self.data
		# возвращает True/False

		if isinstance(files, str):
			files = [files]

		to_validate = []
		for fn in files:
			if fn.startswith('*'):
				fn = fn[1:]
				to_validate.append(fn)

			if fn.endswith('/'):
				self.data[fn] = os.listdir('res/' + fn)
			else:
				self.data[fn] = self.file.load(fn + '.txt')

		result_done = True
		if to_validate:
			for v in to_validate:
				if not self.data[v]:
					result_done = False
					self.log('нет данных в res/' + v, '!')

		return result_done

	def w_text_log(self):

		if not self.cnf.getbool('text_log'):
			return

		rows = []

		fp = open(self.parent.logname, 'ab')

		while True:

			if self.parent.threadsStopped.is_set():
				#сливаем все оставшиеся в файл
				fp.write(bytearray('\n'.join(rows), 'utf-8'))
				break

			row = False
			try:
				row = self.parent.log_queue.get(True, 1)
			except queue.Empty:
				#time.sleep(1)
				break

			if row:
				rows.append(row)

			if len(rows) > self.cnf.getint('flush_log'):
				fp.write(bytearray('\n'.join(rows), 'utf-8'))
				rows = []

		# закрываем лог-файл
		fp.close()

	def w_sqlite_cache(self):
		# поток, пополняющий sqlite_storage для других потоков

		sqlite = asSqlite(self.cnf.getbool('sqldebug'), parent=self.parent, dbpath=self.parent.db_path, lightload=True)

		# подкачиваем: proxy, recipients(invalid,spammed,sent_failed)
		# получателей качать если не спам по контакт-листу
		if self.cnf.getint('spam_mode') == 0:
			self.parent.sqlite_storage['recipients'] = []

		while True:

			if self.parent.threadsStopped.is_set():
				break

			# прокси, если обычные
			#if not self.cnf.getbool('proxy_backconnect'):
			#~ if not self.cnf.getbool('proxy_src_url'):
				#~ self.parent.sqlite_storage['proxy'] = sqlite.getlist('proxy', what='data, type')#limit=300,

			if self.cnf.getint('spam_mode') == 1:
				to_cache = sqlite.getlist('recipients', where2='invalid = 0 and spammed = 0 and sent_failed = 0', limit=300)

				self.parent.sqlite_storage['recipients'] = to_cache

			for i in range(3):
				if self.parent.threadsStopped.is_set():
					break
				time.sleep(1)

	def w_sqlite(self):
		# обработка всех поступивших команд sqlite

		try:
			self._process_requests(self.parent)
		except Exception as e:

			e = str(e)
			tb = sys.exc_info()[2]
			if tb:
				tb = ''.join(traceback.format_tb(tb))
			else:
				tb = 'Not detected'

			with open('SQLITE_PROCESS_CRASHED', 'w') as fp:
				fp.write(e +' --- '+ tb)

			with open('engine/cmd.txt', 'w') as fp:
				print('SQLITE_PROCESS_CRASHED! Stopping..')
				fp.write('stop')

	# old w_sqlite. run from w_sqlite
	def _process_requests(self, parent):

		#self.parent = parent

		sqlite = asSqlite(self.cnf.getbool('sqldebug'), parent=self.parent, dbpath=self.parent.db_path, lightload=True, complex_commits=False,name='w_sqlite')

		# DBG
		#fp = open('w_sqlite_stat', 'w')
		#fp = open('/dev/null', 'w')
		#got = finished = 0

		while True:

			if self.parent.threadsStopped.is_set() or os.stat('engine/cmd.txt')[6]:
				# до этого не доходит, если вырубить скрипт Escape
				sqlite.conn.commit()
				sys.exit(0)
				break

			# DBG
			#total = str(self.parent.queue.tasks_for_main.qsize())
			#t = str(time.time())

			try:
				data = self.parent.queue.tasks_for_main.get(True, 1)
			except queue.Empty:
				continue

			# DBG
			#got += 1
			#t0 = Decimal(time.time())

			if isinstance(data, dict):
				if 'sql' in data:
					cmd = data['sql']

					if cmd == 'rm':
						sqlite.rm(data['table'], data['rows'])

					if cmd == 'unique':
						sqlite.unique(data['table'], data['column'])

					if cmd == 'clear':
						sqlite.clear(data['table'])

					if cmd == 'upd':
						sqlite.upd(data['table'], data['data'], data['action'])

					if cmd == 'add':
						cols = data['cols'] if 'cols' in data else None
						split = data['split'] if 'split' in data else None

						sqlite.add(data['table'], data['rows'], cols, split)

					if cmd == 'getlist':
						# hash для уникальности данных при многопотоке
						self.parent.sqlite_storage[data['hash']] = sqlite.getlist(data['table'], data['unique'], data['valid'], data['where2'], data['limit'], data['what'])

					if cmd == 'exec':
						sql_id = data['sql_id']
						sql = data['sql_cmd']

						data = sqlite.exec(sql)

						if sql.lower().startswith('select'):
							if data is False:
								print('FALSE!!! as_bot sql exec: ' + sql)
								self.parent.sqlite_storage[sql_id] = False
							else:
								self.parent.sqlite_storage[sql_id] = data

			# DBG
			#~ finished += 1
			#~ t1 = Decimal(time.time())
			#~ fp.write('{0}\ttotal {1}\tfinished {2}\tgot {3}\tchanges {4}\texectime {5}\n'.format(t, total, finished, got, sqlite.changes, str(t1-t0)))
			#~ fp.flush()
#~
		#~ fp.close()

	def w_sqlite_OLD(self):
		# обработка всех поступивших команд sqlite

		sqlite = asSqlite(self.cnf.getbool('sqldebug'), parent=self.parent, dbpath=self.parent.db_path, lightload=True)

		while True:

			if self.parent.threadsStopped.is_set():
				break

			try:
				# блочить поток до 30 сек, пока не найдем задачу
				data = self.parent.queue.tasks_for_main.get(True, 1) # True, 30
			except queue.Empty:
				#time.sleep(2)
				continue

			#print('<p>NEW COMMAND', data, '</p>')
			try:
				if isinstance(data, dict):
					if 'sql' in data:
						cmd = data['sql']

						if cmd == 'rm':
							sqlite.rm(data['table'], data['rows'])

						if cmd == 'unique':
							sqlite.unique(data['table'], data['column'])

						if cmd == 'clear':
							sqlite.clear(data['table'])

						if cmd == 'upd':
							print('UPDATING {0}, data: {1}, action: {2}'.format(data['table'], data['data'], data['action']))
							sqlite.upd(data['table'], data['data'], data['action'])

						if cmd == 'add':
							cols = data['cols'] if 'cols' in data else None
							split = data['split'] if 'split' in data else None
							sqlite.add(data['table'], data['rows'], cols, split)

						if cmd == 'getlist':
							# hash для уникальности данных при многопотоке
							self.parent.sqlite_storage[data['hash']] = sqlite.getlist(data['table'], data['unique'], data['valid'], data['where2'], data['limit'], data['what'])

						if cmd == 'exec':
							sql_id = data['sql_id']
							sql = data['sql_cmd']
							#condition = data.get('sql_condition', False)

							#print('bot: CONDITION', condition)
							data = sqlite.exec(sql)

							if sql.lower().startswith('select'):
								if data is False:
									print('FALSE!!! as_bot sql exec: ' + sql)
									self.parent.sqlite_storage[sql_id] = False
								else:
									#print('bot: ADDED ' + sql_id)
									self.parent.sqlite_storage[sql_id] = data

							#~ if condition:
								#~ print('bot: NOTIFY')
#~
								#~ with condition:
									#~ condition.release()

			finally:
			# не надо
				self.queue.tasks_for_main.task_done()

	def w_update_accs(self, url, delay):

		while True:

			if self.parent.threadsStopped.is_set():
				break

			for i in range(delay * 60):
				if self.parent.threadsStopped.is_set():
					break
				time.sleep(1)

			self.log(('обновляем список акков с урла ', url))
			accs = self.parse_url(url)

			if isinstance(accs, list) and len(accs):
				sql = asSqlite(self.cnf.getbool('sqldebug'), parent=self.parent, dbpath=self.parent.db_path, lightload=True)
				sql.add('accounts', accs)

			if delay == 0:
				break

	def parse_url(self, url):
		''' парсит акки с урла url, возвращает список '''

		page = self.parent.http.get(url)

		if self.parent.http.last_error:
			self.log(('ошибка скачивания по ссылке: ', self.parent.http.last_error), '!')
			return []

		accs = list(set(page.split('\n')))

		if not len(accs):
			self.log('не найдены акки по ссылке', '!')
			return []
		else:
			self.log(['загружено {0} акков', len(accs)], '+')

		return accs

	def w_check_cmd(self):

		while True:

			if self.parent.threadsStopped.is_set() or os.stat('engine/cmd.txt')[6]:

				self.log('останавливаем скрипт и потоки', '!')

				self.parent.threadsStopped.set()
				self.threads.stopHook()
				print('<script>clearInterval(window.timerInt)</script>')
				sys.exit(0)
				break
			# обновляем счетчики

			if self.parent.cnf_id:
				path = 'configs/%s/counters' % self.parent.cnf_id
			else:
				path = 'counters'

			text = ''
			for key, counter in self.parent.counters.items():
				#name, cdata
				text += '{0}:{1}\n'.format(key, counter['value'])

			if self.cnf.getbool('proxy'):
				total_proxies = len(self.parent.sqlite_storage['proxy'])
				if total_proxies and self.parent.proxy.last_download_count:
					percent = int(total_proxies / (self.parent.proxy.last_download_count/100))
				else:
					percent = 0

				text += 'proxies:{0} ({1}%)\n'.format(total_proxies, percent)
			else:
				text += 'proxies:Off\n'

			with open(path, 'wb') as fp:
				fp.write(bytearray(text, 'utf-8'))

			time.sleep(1)

























# Fucking bugZ-bunny killed with BigFuckingGun! HEad Shot AAHHHAHHHHAH
