import configparser

class asCnf():

	config = None
	options = None
	cached = {}
	lang_rows = {}

	def __init__(self, parent=None, options=[], cnf_path='config.ini'):

		self.path = cnf_path
		self.options = options

		self.config = configparser.RawConfigParser(inline_comment_prefixes=(';'))
		try:
			self.config.read(self.path, 'utf-8')
		except Exception as t:
			print(t)

		#for option, value in self.config.items('default'):
		#	print(option, value)

		#self.updateGui()

	def reset(self):
		for option, value in self.config.items('default'):
			self.config.set('main', option, value)
		self.updateGui()

	def chk(self, option):

		return True if option in self.config['main'] else False

	# not implemented in web-version
	def updateGui(self):
		''' not implemented in web-version  '''
		for block in self.options:
			if block == 'checkbox':
				for option in self.options[block]:
					val = self.getbool(option)
					#print('as_cnf: %s=%s' % (option, str(val)))
					self.gui.chkbox(option, val)
			elif block == 'lineedit':
				for option in self.options[block]:
					val = self.get(option)
					#print('as_cnf: %s=%s' % (option, val))
					self.gui.lineEdit(option, 'set', val)
			elif block == 'combobox':
				for option in self.options[block]:
					val = self.getint(option)
					#print('as_cnf: %s=%s' % (option, str(val)))
					self.gui.combo(option, 'set', val)
			elif block == 'radio':
				for option in self.options[block]:
					val = self.getbool(option)
					#print('as_cnf: %s=%s' % (option, str(val)))
					self.gui.radio(option, val)

	def print(self):
		'''print all settings'''
		for opt, name in self.config.items('main'):
			print(opt+'='+name)

	def load(self, file):
		res = self.config.read(file)
		if not len(res):
			print('конфиг', file, 'пустой!')

	def get(self, option):

		if option in self.cached:
			return self.cached[option]
		else:
			self.cached[option] = self.config.get('main', option).strip()
			return self.cached[option]

	def getint(self, option):

		if option in self.cached:
			return int(self.cached[option])
		else:
			self.cached[option] = self.config.getint('main', option)
			return self.cached[option]

	def getbool(self, option):

		if option in self.cached:
			return bool(self.cached[option])
		else:
			self.cached[option] = self.config.getboolean('main', option)
			return self.cached[option]

	def set(self, option, value):
		try:
			self.config.set('main', option, str(value))
		except configparser.NoOptionError:
			print('опция', option, 'не найдена!')
			return None

	def save(self):
		''' not implemented in web-version  '''
		#get from gui
		for block in self.options:
			if block == 'checkbox':
				for option in self.options[block]:
					val = self.gui.chkbox(option, 'get')
					self.set(option, val)
			elif block == 'lineedit':
				for option in self.options[block]:
					val = self.gui.lineEdit(option, 'get')
					self.set(option, val)
			elif block == 'combobox':
				for option in self.options[block]:
					val = self.gui.combo(option, 'get')
					self.set(option, val)
			elif block == 'radio':
				for option in self.options[block]:
					val = self.gui.radio(option, 'get')
					self.set(option, val)

		#save to file
		with open(self.path, 'w') as configfile:
			self.config.write(configfile)

		self.config.read(self.path)
		self.updateGui()
