import sqlite3, itertools, os, time
from urllib.parse import quote, unquote
from threading import RLock, Condition

# 01 августа добавил RLock
class asSqlite():

	debug = False
	conn = None
	tables = []
	changes = 0 # кол-во требуемых коммитов для _commit()
	complex_commits = False # делать коммиты пачками
	table_cols = {} # строки для вставки в таблицы, типа 'null, "%s", 0'
	iterators = {} # для getnext
	lock = RLock()

	def __init__(self, debug=False, parent=None, dbpath='./data.db', lightload=False, complex_commits=False, name=''):

		self.parent = parent
		self.complex_commits = complex_commits

		if debug:
			self.debug = True

		self.conn = sqlite3.connect(dbpath, isolation_level='IMMEDIATE')
		self.cursor = self.conn.cursor()

		if lightload:
			return

		if hasattr(self.parent, 'sqlite_storage'):
			# кешируем таблицы

			cache_tables = self.exec('select tablename from _descriptions where cacheable = 1')
			cache_tables = [row[0] for row in cache_tables]

			for cache_table in cache_tables:

				if cache_table == 'proxy':
					continue # кешируем в as_proxy.load()

				self.parent.data[cache_table] = self.getlist(cache_table, what='data')

		if self.parent and self.parent.sqltables:
			self.tables = self.parent.sqltables
			self.table_cols = self.parent.sqltables_cols

		if not self.tables:
			self.tables = self.exec('select name from sqlite_master where type in ("table", "view") and name not like ("sqlite_%")')
			self.tables = [row[0] for row in self.tables] # if not row[0].startswith('_')

			for table in self.tables:
				cols = self.exec("PRAGMA table_info('%s');" % table)
				if not cols:
					continue
				string = ''
				for col in cols:
					col_name = col[1]
					def_val = col[4]

					if col_name == 'data':
						string += "'%s', "
					elif def_val is None:
						string += 'null, '
					elif def_val == '':
						string += "'', "
					else:
						string += str(def_val) + ', '

				if string.endswith(', '):
					string = string[:-2]

				self.table_cols[table] = string

			if self.parent:
				self.parent.sqltables = self.tables
				self.parent.sqltables_cols = self.table_cols

	def dbg(self, sql):

		if not self.debug:
			return

		print(sql)
		with open('SQLITE_DEBUG', 'a') as  fp:
			fp.write(sql + '\n')

	def _commit(self):

		if not self.complex_commits:
			self.conn.commit()
			return

		self.changes += 1

		if self.changes > 10:
			self.changes = 0
			self.conn.commit()

	def exec(self, sql):
		''' возвращает список всех таблиц с данными '''

		if self.debug:
			self.dbg('sql: exec: %s' % sql)

		commit = True
		if 'NO_COMMIT' in sql: # нельзя в многопотоке
			commit = False
			sql = sql.replace('NO_COMMIT', '')

		with self.lock:
			try:
				#cursor = self.conn.cursor()
				tables = self.cursor.execute(sql.strip()).fetchall()
				#cursor.close()
			except Exception as ex:
				self.dbg('sql: exec: Exception: %s' % str(ex))
				return False

			if not sql.lower().startswith('select') and commit:
				self._commit()

		return tables

	def get(self, table):
		''' достает данные из первой строки таблицы, возвращает строку '''

		sql = 'select data from `%s`' % table

		if self.debug:
			self.dbg('sql: get: %s' % sql)

		with self.lock:
			#cursor = self.conn.cursor()
			data = self.cursor.execute(sql).fetchall()
			#cursor.close()

		return unquote(data[0][0])

	def getlist(self, table, unique=True, valid=False, where2='', limit=None, what='data'):
		''' достает все строки из таблицы, возвращает список '''

		if valid:
			sql = 'select %s from `%s` where invalid = 0' % (what, table)
			if where2.strip():
				sql += ' and %s' % where2
		else:
			sql = 'select %s from `%s`' % (what, table)
			if where2.strip():
				sql += ' where %s' % where2

		if limit:
			sql += ' limit %d' % int(limit)

		if self.debug:
			self.dbg('sql: getlist: %s' % sql)

		with self.lock:

			try:
				cursor = self.conn.cursor()

				if what == 'data, type':
					rows = [[unquote(row[0]), row[1]] for row in cursor.execute(sql).fetchall() if row[0]]
				else:
					rows = [unquote(row[0]) for row in cursor.execute(sql).fetchall() if row[0]]
				cursor.close()

				if unique and len(rows) > 1:
					rows = list(set(rows))

			except Exception as ex:
				self.dbg('can"t execute sql: %s, %s' % (sql, ex))
				return []

		if self.debug:
			self.dbg('sql: getlist: got %d rows from %s' % (len(rows), table))

		return rows

	def sz(self, table):
		return self.size(table)

	def size(self, table):
		''' возвращает количество строк в таблице table '''

		sql = 'select COUNT(data) from %s where data != "" and data != " "' % table

		if self.debug:
			self.dbg('sql: size: %s' % sql)

		with self.lock:
			#cursor = self.conn.cursor()
			data = self.cursor.execute(sql).fetchall()
			#cursor.close()

		return int(data[0][0])

	def rm(self, table, rows):
		''' удаляет строку row из таблицы table '''

		if isinstance(rows, str):
			rows = rows.strip().split('\n')

		for row in rows:
			sql = 'delete from `%s` where data="%s"' % (table, quote(row.strip()))

			if self.debug:
				self.dbg('sql: rm: %s' % sql)

			with self.lock:
				#cursor = self.conn.cursor()
				self.cursor.execute(sql)
				#cursor.close()

				self._commit()

	def upd(self, table, data, action):
		''' обновляет строку с data в таблице table. action - стандартные обновления

			sql.upd('accounts', 'vasya', 'invalid')
			sql.upd('accounts', 'vasya', 'empty_cl = 1')
		'''

		known_flags = ['invalid', 'valid', 'unknown_service', 'empty_cl', 'not_confirmed']

		if action in known_flags:
			upd = action + ' = 1'
		else:
			upd = action

		sql = 'update `%s` set %s where data = "%s"' % (table, upd, quote(data))

		if self.debug:
			self.dbg('sql: update: %s' % sql)

		with self.lock:
			#cursor = self.conn.cursor()
			self.cursor.execute(sql)
			#cursor.close()
			self._commit()

	def unique(self, table, column=None):
		''' удаляет дубли из таблицы '''

		cvars = (table, table)
		sql = 'DELETE FROM `%s` WHERE auto_id NOT IN (SELECT MAX(auto_id) FROM `%s` GROUP BY data)' % cvars

		if column:
			cvars = (table, table, column)
			sql = 'DELETE FROM `%s` WHERE ROWID NOT IN (SELECT MIN(ROWID) FROM `%s` GROUP BY %s)' % cvars

		if self.debug:
			self.dbg('sql: unique: %s' % sql)

		with self.lock:
			#cursor = self.conn.cursor()
			try:
				self.cursor.execute(sql)
				self._commit()
			except sqlite3.OperationalError:

				# очистка дублей для таблиц, где нет auto_id
				data = self.getlist(table)
				self.clear(table)
				self.add(table, data)

	def clear(self, table):
		''' очищает таблицу table '''

		sql = 'delete from `%s`' % table

		if self.debug:
			self.dbg('sql: clear: %s' % sql)

		with self.lock:
			#cursor = self.conn.cursor()
			self.cursor.execute(sql)
			#cursor.close()
			self._commit()

	def mark_invalid(self, table, rows):
		''' отмечает строку как invalid '''

		if isinstance(rows, str):
			rows = rows.strip().split('\n')

		for row in rows:
			sql = 'update `%s` set invalid = 1 where data = "%s"' % (table, row.strip())

			if self.debug:
				self.dbg('sql: mark_invalid: %s' % sql)

			with self.lock:
				#cursor = self.conn.cursor()
				self.cursor.execute(sql)
				#cursor.close()
				self._commit()

	def add(self, table, rows, cols=None, split=True):
		''' добавляет строку row в конец таблицы table '''

		#self.parent.log(', '.join(rows))
		if not rows:
			self.dbg('sql: add: 0 ROWS')

		if not cols:
			cols = self.table_cols[table]

		if isinstance(rows, int):
			rows = str(rows)

		if isinstance(rows, str):
			if split:
				rows = rows.strip().split('\n')
			else:
				rows = [rows]

		if len(rows) == 1:
			with self.lock:
				sql = 'insert or ignore into `%s` values(%s)' % (table, cols)

				#cursor = self.conn.cursor()
				self.cursor.execute(sql % quote(rows[0]))
				#cursor.close()

				self._commit()
			return

		sql = 'insert or ignore into `%s` values(%s)' % (table, cols.replace("'%s'", '?').replace('"%s"', '?'))

		if self.debug:
			self.dbg('sql: add: %s' % sql)

		def generator(rows):
			for row in rows:
				yield (quote(row.strip()),)

		with self.lock:
			#cursor = self.conn.cursor()
			self.cursor.executemany(sql, generator(rows))
			#cursor.close()
			self._commit()

	def save(self, table, rows, cols=None, split=True):
		''' сохраняет список lines построчно в таблицу table '''

		self.clear(table)
		self.add(table, rows, cols=cols, split=split)

	def getrand(self, table, what='data', where=None):
		''' достает рандом строку из таблицы table. возвращает set с полями '''

		if where:
			where_t = 'where ' + where
		else:
			where_t = ''

		sql = 'select %s from `%s` %s order by RANDOM() limit 1' % (what, table, where_t)

		if self.debug:
			self.dbg('sql: getrand: %s' % sql)

		try:
			with self.lock:
				row = self.cursor.execute(sql).fetchall()[0]
		except Exception as ex:
			self.dbg('sql: exception "%s" with SQL: %s' % (ex, sql))
			return ''

		row = [self._unquote_b(col) for col in row]

		if len(row) == 1:
			return row[0]

		return row

	def _unquote_b(self, text):
		return unquote(text) if isinstance(text, str) else text

	def getnext(self, table):

		if table in self.iterators:
			row = next(self.iterators[table])
		else:
			with self.lock:
				#cursor = self.conn.cursor()
				res = self.cursor.execute("SELECT * FROM `%s`" % table).fetchall()
				#cursor.close()

			self.iterators[table] = itertools.cycle(res)
			row = next(self.iterators[table])

		row = [self._unquote_b(col) for col in row]

		if len(row) == 1:
			return row[0]

		return row

	#~ def test(self):
#~
#~
		#~ import sqlite3, sys
#~
		#~ conn = sqlite3.connect('./data.db', isolation_level='IMMEDIATE')
		#~ cursor = conn.cursor()
#~
		#~ s = 'select * from accounts'
		#~ d = 'delete from accounts'
#~
		#~ t = cursor.execute(d).fetchall()
		#~ print(t)
		#~ sys.exit(0)


# "sqlite отложенный" - обертка, чтобы обращаться к базе из потоков
# все запросы помещаются в очередь, а поток include/as_bot.py w_sqlite
# - разбирает эти запросы и какбэ выполняет

class asSqlite_deferred():

	debug = False

	def __init__(self, debug=False, parent=None, dbpath='./data.db', lightload=False):
		self.parent = parent
		self.tid = parent.tid

		if debug:
			self.debug = True

	def getrand(self, table, what='data', where='', rm=False):
		from engine.as_hlp import hlp

		if table not in self.parent.sqlite_storage:
			sql = 'select %s from `%s` %s order by RANDOM() limit 1' % (what, table, where)
			try:
				string = self.exec(sql, save=rm)[0][0]
			except TypeError:
				return False
			except IndexError:
				return False

			return hlp.unquote(string)

			#return 'as_sqlite_deferred.getrand: table "%s" not found in sqlite_storage' % table

		string = hlp.rand(self.parent.sqlite_storage[table])

		if rm:
			if string in self.parent.sqlite_storage[table]:
				del self.parent.sqlite_storage[table][string]

		return string

	def sz(self, table):
		res = self.exec('select COUNT(data) from %s where data != "" and data != " "' % table)
		return res[0][0]

	def rm(self, table, rows):
		self.parent.queue.addTask({'sql': 'rm', 'table': table, 'rows': rows})

	def unique(self, table, column=None):
		self.parent.queue.addTask({'sql': 'unique', 'table': table, 'column': column})

	def upd(self, table, data, action):
		self.parent.queue.addTask({'sql': 'upd', 'table': table, 'data': data, 'action': action})

	def add(self, table, rows, cols=None, split=True):
		self.parent.queue.addTask({'sql': 'add', 'table': table, 'rows': rows, 'cols': cols, 'split': split})

	def getlist(self, table, unique=True, valid=False, where2='', limit=None, what='data'):

		try:
			from engine.as_hlp import hlp
		except:
			self.dbg('as_sqlite: cannot import hlp')

		hash_t = hlp.md5(table + str(hlp.time(full=True)))

		self.parent.queue.addTask({'sql': 'getlist', 'table': table, 'unique': unique, 'valid': valid, 'where2': where2, 'limit': limit, 'what': what, 'hash': hash_t})

		for i in range(5):

			if self.parent.threadsStopped.is_set():
				break

			time.sleep(i)

			if hash_t in self.parent.sqlite_storage.keys():

				data = self.parent.sqlite_storage[hash_t]
				del self.parent.sqlite_storage[hash_t]
				return data

		#print('as_sqlite_def.getlist Fail with table: %s' % table)
		return ''

	def exec(self, sql, save=True):
		import time

		sql_id = '{0}.{1}.{2}'.format(self.tid, sql[:30], str(time.time()))

		self.parent.queue.addTask({'sql': 'exec', 'sql_cmd': sql, 'sql_id': sql_id})

		if not sql.lower().startswith('select'):
			return

		i = -1
		while True:
			i += 1

			if self.parent.threadsStopped.is_set():
				return

			if sql_id in self.parent.sqlite_storage:

				data = self.parent.sqlite_storage[sql_id]
				if not save and sql_id in self.parent.sqlite_storage:
					del self.parent.sqlite_storage[sql_id]

				return data

			time.sleep(1)

		raise Exception('as_sqlite_def.exec Fail with SQL: %s' % sql)
		return False


#~ def main():
	#~ sql = asSqlite()
	#~ sql.debug = True

	#for i in range(10):
	#	sql.mark_invalid('accounts', i)

	#sql.unique('accounts')
	#
	#sql.rm('accounts', '1')

	#~ accs = sql.getlist('accounts')
	#~ print('>>', accs)

	#sql.clear('accounts')




if __name__ == '__main__':
	main()


