import sys, traceback, queue
#from PyQt4 import QtGui, QtCore
# install from http://pypi.python.org/pypi/futures/2.1.2
# doc in http://packages.python.org/futures/
from concurrent.futures import ThreadPoolExecutor, wait
#QtCore.QObject
from engine.as_hlp import hlp
from engine.as_sqlite import asSqlite_deferred

#import threading
#threading.stack_size(64*1024)
#~ import resource
#~ resource.setrlimit(resource.RLIMIT_STACK, (resource.RLIM_INFINITY, resource.RLIM_INFINITY))

#sys.setrecursionlimit(300)


class asThreads():

	jobs = 0
	stop = False
	pending = set()
	executor = None
	worker = None #  назначается в Bot
	callback_on_finish = None # вызвать после завершения всех потоков
	callback_on_thread_result = None # вызвать после завершения потока
	infinity_work = False #  работать бесконечно -- не посылать сигнал к завершению, не писать set_stop

	def __init__(self, parent):
		self.parent = parent
		self.hlp = self.parent.hlp
		self.log = self.parent.log
		self.cnf = self.parent.cnf
		self.sqlite = self.parent.sqlite

	def set_worker(self, worker):
		self.worker = worker

	def process_table (self, threads_n, table_name, worker, where_params):

		if not where_params.startswith('where '):
			where_params = 'where ' + where_params

		self.where_params = where_params
		self.table_name = table_name
		# ставим воркер, разбирающий эту очередь
		self.set_worker(self.w_work_manager_table)
		# ставим свой воркер уже для менеджера
		self.manager_worker = worker
		# запускаем Н потоков
		self.run_n(threads_n)

	def process(self, threads_n, data, worker):

		# создаем очередь из data
		self.rows = queue.Queue()
		for row in data:
			self.rows.put(row)

		# ставим воркер, разбирающий эту очередь
		self.set_worker(self.w_work_manager)
		# ставим свой воркер уже для менеджера
		self.manager_worker = worker
		# запускаем Н потоков
		self.run_n(threads_n)

	def run_n(self, maxt, worker=''):
		'''
			запускает max одинаковых потоков.
			например, для реггера
		'''
		if worker:
			self.worker = worker

		if not maxt:
			self.log('0 threads finished')
			return

		self.run(['' for i in range(maxt)])

	def run(self, jobs, attr=[]):
		'''
			jobs - список заданий. 1 задание - 1 поток.
			attr - список параметров, одинаковых для каждого задания
		'''

		self.parent.gui.lbl('threads', 'set', 0)

		self.jobs = len(jobs)

		mt_cnf = self.cnf.getint('threads')
		mt_real = self.jobs

		if mt_real < mt_cnf:
			mt_cnf = mt_real

		self.executor = ThreadPoolExecutor(max_workers=mt_cnf)

		for i, job in enumerate(jobs):

			future = self.executor.submit(self._worker, [i, job, attr])
			future.add_done_callback(self.process_result)
			self.pending.add(future)

	def w_work_manager_table(self, param):

		# разбирает из таблицы, таблица должна иметь флаг processed
		tid, n, args = param

		sqlite = asSqlite_deferred(self.cnf.getbool('sqldebug'), parent=self.parent)

		acc = sqlite.getrand(self.table_name, where=self.where_params)

		if acc is False:
			self.log('строки закончились', '-', tid=tid)
			acc = False

		processed = 0
		while acc is not False:

			if processed == self.cnf.getint('process_limit'):
				self.log('лимит {0} в настройках, завершаем работу потока'.format(self.cnf.getint('process_limit')))
				break

			# обрабатываем
			res = self.manager_worker(acc, tid)
			processed += 1

			if res == 'skip':
				self.log('прерываем рассылку с этого потока', '-', tid=tid)
				break
			elif res == 'repeat':
				self.log('повторяем обработку строки ' + acc, '*', tid=tid)
				continue
			else:
				self.log('строка обработана', '+', tid=tid)
				sqlite.upd(self.table_name, acc, 'processed=1')

			self.log('берем новую строку', tid=tid)
			acc = sqlite.getrand(self.table_name, where=self.where_params)
			if not acc:
				self.log('строки закончились', '-', tid=tid)
				acc = False
			else:
				self.log('следующая строка <b>{0}</b>'.format(acc), '*', tid=tid)

			if self.parent.threadsStopped.is_set():
				self.log('включен флаг threadsStopped, прерываем работу потока', '!', tid=tid)
				break

		self.log('цикл потока завершен', '+', tid=tid)

	def w_work_manager(self, param):

		tid, n, args = param

		try:
			acc = self.rows.get(True, 1)
		except queue.Empty:
			self.log('строки закончились', '-', tid=tid)
			acc = False

		processed = 0
		while acc is not False:

			if processed == self.cnf.getint('process_limit'):
				self.log('лимит {0} в настройках, завершаем работу потока'.format(self.cnf.getint('process_limit')))
				break

			# обрабатываем
			res = self.manager_worker(acc, tid)
			processed += 1

			if res == 'skip':
				self.log('прерываем рассылку с этого потока', '-', tid=tid)
				break
			elif res == 'repeat':
				self.log('повторяем обработку строки ' + acc, '*', tid=tid)
				continue
			else:
				self.log('строка обработана', '+', tid=tid)

			try:
				self.log('берем новую строку', tid=tid)
				acc = self.rows.get(True, 1)
			except queue.Empty:
				self.log('строки закончились', '-', tid=tid)
				acc = False

			if acc:
				self.log('следующая строка <b>{0}</b>'.format(acc), '*', tid=tid)

			if self.parent.threadsStopped.is_set():
				self.log('включен флаг threadsStopped, прерываем работу потока', '-', tid=tid)
				break

		self.log('цикл потока завершен', '+', tid=tid)

	def _worker(self, args):

		tid, email, param = args

		event = self.parent.queue.events['stop_right_there']
		if event.is_set():
			event.clear()
			return

		#self.parent.queue.addTask('thread_start')
		self.hlp.cnt('threads')
		try:
			self.worker(args)
		except Exception as e:
			self.log('Thread exception in threads._worker: %s' % e, '!', tid=tid)
			print(traceback.print_exc(file=sys.stdout))

			e = str(e)
			tb = sys.exc_info()[2]
			if tb:
				tb = ''.join(traceback.format_tb(tb))
			else:
				tb = 'Not detected'

			with open('SQLITE_THREAD_'+str(tid)+'_CRASHED', 'w') as fp:
				fp.write(e +' --- '+ tb)

	def process_result(self, future):
		if future.cancelled():
			res = ''
			self.parent.queue.addTask(('thread_done'))
		else:
			pass


		if self.callback_on_thread_result:
			self.callback_on_thread_result()

		res = future.result(timeout=60)
		self.pending.discard(future)
		self.jobs -= 1
		self.hlp.cnt('threads', 'dec')

		if self.jobs == 0:
			self.log('закончили работу', '+')

			#self.parent.queue.tasks_for_main.join() - ждет пока вся очередь будет обработана до конца

			if not self.infinity_work:
				# чтобы остановить поток автообновления прокси и тп
				self.parent.threadsStopped.set()

				print('<script>stop_script()</script>')

			#hlp.stop()
			self.callback_on_thread_result = None
			if self.callback_on_finish:
				callback = self.callback_on_finish
				self.callback_on_finish = None
				callback()

	def close(self, event):

		print('ЗАКРЫВАЕМ ОКНО')
		if self.pending == False:
			return

		event.accept()
		self.stopHook()
		#self.executor.shutdown()

	def set_callback_on_thread_finish(self, callback):
		'''вызывать при завершении каждого потока'''
		self.callback_on_thread_result = callback

	def set_callback_on_finish(self, callback):
		'''вызвать при завершении всех потоков'''
		self.callback_on_finish = callback

	def stopHook(self):
		# stop button

		for f in set(self.pending):
			if not f.done() and not f.cancelled():
				#self.parent.queue.addTask(('thread_done', 'cancelled'))
				f.cancel()

	#~ def shutdown(self):
		#~ self.executor.shutdown()

	def show(self):
		''' распечатка потоков по button_test '''
		data = wait(self.pending)
		print(data)
















