import os, sys, random

def _print(text):
	sys.stdout.buffer.write(text.encode('utf8'))
	sys.stdout.flush()

class Config(object):

	current_section = ''
	sections = {}
	odd = False
	blocks = {}
	lang_rows = {}

	def __init__(self, path):
		self.path = path
		pass

	def parseLang(self):

		if os.path.exists('res/lang'):
			for r in open('res/lang', 'rb'):
				row = r.decode('utf-8', 'ignore').strip()
				if not row.startswith('#') and row.strip() and '|' in row:
					src, eng = row.split('|')
					self.lang_rows[src] = eng

	def trans(self, text):
		return self.lang_rows.get(text, text)

	def _processRow(self, line):
		'''разделяет строки и секции'''
		if self.current_section not in self.sections:
			self.sections[self.current_section] = []

		line_data = {'name': '', 'comments': '', 'type': '', 'value': ''}

		parts = line.split(';')

		try:
			key, value = parts[0].split('=', 1)
		except:
			key = parts[0].strip()
			value = ''
		try:
			line_data['name'] = key.strip()
			line_data['value'] = value.strip()

			if len(parts) == 3:
				line_data['type'] = parts[1].strip()
				comm = parts[2].strip()
				#comm = self.lang_rows.get(comm, comm)

				line_data['comments'] = comm
			else:
				line_data['type'] = ''

				comm = parts[1].strip()
				#comm = self.lang_rows.get(comm, comm)

				line_data['comments'] = comm
		except:
			pass

		#_print('append: ' + line_data['name'] + '<br />')
		self.sections[self.current_section].append(line_data)

	def _formatRow(self, line):
		'''форматирует строку, разбивает на части'''

		parts = line.split(';')

		# ключ = значение  БЕЗ КОММЕНТА
		if len(parts) == 1:
			subparts = line.split('=', 1)

			# ключ =
			if len(subparts) == 1:
				key = subparts[0].strip()
				val = ''

			# ключ = значение
			if len(subparts) == 2:
				key = subparts[0].strip()
				val = subparts[1].strip()

			if self._is_bool(val):
				return self._makeChkbox(key, val)
			else:
				return self._makeTextInput(key, val)

		# ключ = значение ; коммент
		if len(parts) == 2:
			subparts = parts[0].strip().split('=', 1)
			comment = parts[1].strip()

			# ключ =
			if len(subparts) == 1:
				key = subparts[0].strip()
				val = ''

			# ключ = значение
			if len(subparts) == 2:
				key = subparts[0].strip()
				val = subparts[1].strip()

			if self._is_bool(val):
				return self._makeChkbox(key, val, comment)
			else:
				return self._makeTextInput(key, val, comment)

		# ключ = значение ; дефолт ; коммент
		if len(parts) == 3:
			subparts = parts[0].split('=', 1)

			# ключ =
			if len(subparts) == 1:
				key = subparts[0].strip()
				val = ''

			# ключ = значение
			if len(subparts) == 2:
				key = subparts[0].strip()
				val = subparts[1].strip()

			default = parts[1].strip()
			comment = parts[2].strip()

			if default.startswith('radio') or default.startswith('list'):
				radioparts = default.split('=', 1)[1]
				radiovals = {}
				for part in radioparts.split('/'):
					rkey, rval = part.split('|')
					radiovals[rkey] = rval

				if default.startswith('radio'):
					return self._makeRadios(key, val, radiovals, comment)
				else:
					return self._makeCheckboxes(key, val, radiovals, comment)

			if default.startswith('long'):
				return self._makeTextInput(key, val, comment, longer=True)

		return line

	def _is_bool(self, val):
		if val.lower() == 'true' or val.lower() == 'false':
			return True
		else:
			return False

	# test, myValue, {1:"один"}, коммент
	def _makeCheckboxes(self, name, value, defaults, comment=''):

		res = ''
		# 1|один, 2|два
		for key, val in defaults.items():
			rnd = random.randint(0, 5000)
			checked = "checked" if key in value else ""

			res += '<input type="checkbox" title="%s" id="id3%d" name="%s" value="%s" %s /> \
			<label for="id3%d">%s</label>\n' % (key, rnd, name, key, checked, rnd, val)

		return '%s %s <input type="hidden" name="%s" value="emptycheckbox" />' % (comment, res, name)

	# test, myValue, {1:"один"}, коммент
	def _makeRadios(self, name, value, defaults, comment=''):

		res = ''
		for key, val in defaults.items():
			rnd = random.randint(0, 5000)
			checked = "checked" if key == value else ""

			if self.lang_rows:
				if val in self.lang_rows.keys():
					val = self.lang_rows[val]

			res += '<input type="radio" name="%s" value="%s" id="id1%d" %s> \
			<label for="id1%d">%s</label>\n' % (name, key, rnd, checked, rnd, val)

		return '<input type="hidden" name="%s" value="emptyradio" /> %s %s' % (name, comment, res)

	def _makeTextInput(self, name, val, comment='', longer=False):

		size = 50 if longer else 10

		#onmouseover="this.focus()"
		return '<input type="text" size="%d" title="%s" name="%s" \
		value="%s" /> %s' % (size, name, name, val, comment)

	def _makeChkbox(self, name, val, comment=''):
		val = "checked" if val.lower() == 'true' else ""
		if not comment:
			comment = name

		rnd = random.randint(0, 5000)
		return '<input type="hidden" name="%s" value="checkbox" /><input type="checkbox" title="%s" id="id2%d" name="%s" %s /> \
		<label for="id2%d">%s</label>' % (name, name, rnd, name, val, rnd, comment)

	def loadFile(self):
		'''грузим файл в память'''

		with open(self.path, 'rb') as file:
			for line in file.readlines():
				line = line.decode('utf-8', 'ignore').strip()

				if line.strip() == '[main]':
					continue

				# если это имя секции
				if line.startswith('; '):

					sect = line.replace('; ', '').strip()
					#sect = self.lang_rows.get(sect, sect)
					self.current_section = sect
					self.blocks[self.current_section] = self.current_section
					continue

				self._processRow(line)

	def getBlocks(self):
		return self.blocks

	def postSave(self, data):
		'''сохраняем данные'''

		result_config = '[main]\n'

		for sect, opts in self.sections.items():
			i = 0
			result_config += '; %s\n' % sect

			for opt in opts:

				if opt['name'] in data:
					#_print('%s <br />' % opt['name'])
					newval = data.getvalue(opt['name'])

					# набор радиокнопок
					if isinstance(newval, list) and newval[0] == 'emptyradio':
						newval = newval[1]

					# list чекбоксов
					if isinstance(newval, list) and 'emptycheckbox' in newval:
						self.sections[sect][i]['value'] = '"%s"' % ' '.join([l for l in newval if l != 'emptycheckbox'])

						result_config += '%(name)s = %(value)s ; %(type)s ; %(comments)s\n' % (self.sections[sect][i])
						continue

					# один чекбокс
					if isinstance(newval, list) and newval[0] == 'checkbox':
						newval = 'true'

					# один чекбокс
					if newval == 'checkbox':
						newval = 'false'

					self.sections[sect][i]['value'] = newval

					if self.sections[sect][i]['type']:
						result_config += '%(name)s = %(value)s ; %(type)s ; %(comments)s\n' % (self.sections[sect][i])
					else:
						result_config += '%(name)s = %(value)s ; %(comments)s\n' % (self.sections[sect][i])
				else:
					pass
				i += 1

		with open(self.path, 'wb') as f:
			f.write(result_config.encode('utf-8'))

		#print('<meta http-equiv="refresh" content="0; url=start.pyw" />')
		print('<script>parent.closeSettings()</script>')

	def draw(self):
		'''рисуем все строки конфига на экране'''

		i = 0
		for sect, rows in self.sections.items():

			sect = self.lang_rows.get(sect, sect)

			_print('<fieldset id="test%d"><legend onclick="toggle(%d)">%s</legend>\
			<span id="span_%d" style="display: 1none">\n' % (i, i, sect, i))
			for line in rows:
				self.odd = False if self.odd else True
				odd = 'drk' if self.odd else ''

				comm = self.lang_rows.get(line['comments'], line['comments'])

				if line['type']:
					line = self._formatRow('%s = %s ; %s ; %s' % (line['name'], line['value'], line['type'], comm))
				else:
					line = self._formatRow('%s = %s ; %s' % (line['name'], line['value'], comm))

				_print('<p class="logline %s">%s</p>\n' % (odd, line))
			_print('</span></fieldset>')
			i += 1















